import vtk
from vtk.util.numpy_support import numpy_to_vtk
import os

class MeshReader:
    """ Main class for the unstructured mesh reader. Reads the points position
        and creates the cells object

        :param str filename: Filename for the input mesh

        .. csv-table:: Available parameters
            :header: Attribute, Type, Description
            :widths: 20, 10, 70

            `filename`, str, Filename of the mesh
            `nbPoints`, int, Number of points
            `positions`, list, List of points position
            `nbCells`, int, Number of cells
            `cells`, list, List of cells
            `cellsType`, list, List of cells type
            `nbVertices`, int, Number of vertices
            `vertices`, list, List of cells of type vertices
            `nbLines`, int, Number of lines
            `lines`, list, List of cells of type line
            `nbTriangles`, int, Number of triangles
            `triangles`, list, List of cells of type triangle
            `nbTetras`, int, Number of tetrahedron
            `tetras`, list, List of cells of type tetrahedra
            `nbHexas`, int, Number of hexahedron
            `hexas`, list, List of cells of type hexahedron

    """
    def __init__(self, filename):
        self.filename = filename

        if not os.path.exists(filename):
            raise FileNotFoundError(filename)

        reader = vtk.vtkUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()
        self.output = reader.GetOutput()

        self.positions = list()
        for i in range(self.output.GetNumberOfPoints()):
            self.positions.append(self.output.GetPoint(i))

        self.cells = Cells(self.output)

    def __getattr__(self, att):
        """ Accessor for private attributes

            :param str att: Attribute of the MeshReader object

            .. note::

                More attributes are available with the @property decorator,
                see the table of available attributes above
        """
        s="\n"
        if att=="nbPoints": return self.output.GetNumberOfPoints()
        elif att=="nbCells": return self.output.GetNumberOfCells()
        elif att=="nbVertices": return len(self.vertices)
        elif att=="nbLines": return len(self.lines)
        elif att=="nbTriangles": return len(self.triangles)
        elif att=="nbTetras": return len(self.tetras)
        elif att=="nbHexas": return len(self.hexas)

        for k in dir(self):
            if "_" not in k and not "get" in k:
                s+=k+"\n"
        raise AttributeError("{} is no valid attribute\nHere is a list of the attributes : {}".format(att, s[:-1]))

    @property
    def vertices(self):
        return self.getCells(1)
    @property
    def lines(self):
        return self.getCells(3)
    @property
    def triangles(self):
        return self.getCells(5)
    @property
    def tetras(self):
        return self.getCells(10)
    @property
    def hexas(self):
        return self.getCells(12)

    @property
    def cellsType(self):
        cT = list()
        for i in range(self.nbCells):
            cT.append(self.output.GetCellType(i))
        return cT

    def getCells(self, typ):
        cellsId = []
        if self.output is not None:
            for i in range(self.output.GetNumberOfCells()):
                if self.output.GetCellType(i)==typ:
                    cellsId.append(Cell(cellID=i, cell=self.output.GetCell(i)))
        return cellsId

    def __str__(self):
        """ Provides an easy to read overview of the mesh
        """
        s = "\n   ** Class vtk unstructured grid reader **\n"
        s+= f"Filename : {self.filename}\n"
        s+= f"Points : {self.nbPoints}\n"
        s+= f"Cells : {self.nbCells}\n"
        s+= f"Tetras : {self.nbTetras}\n"
        s+= f"Triangles : {self.nbTriangles}\n"
        s+= f"Vertices : {self.nbVertices}\n" # if self.nbVertices>0:
        s+= f"Lines : {self.nbLines}\n" # if self.nbLines>0:
        s+= f"Hexas : {self.nbHexas}\n" # if self.nbHexas>0:
        return s+"\n"

    def getPointData(self, name):
        """ Accessor to the data array of name `name`

            :param str name: Name of the data array as it is in the vtk file

            :returns: DataArray of the vtk data
        """
        vtkArray = self.output.GetPointData().GetArray(name)
        if vtkArray is None:
            raise AttributeError(f"PointData {name} not found in {self.filename}")
        return DataArray(vtkArray)

    def getCellData(self, name):
        """ Accessor to the data array of name `name`

            :param str name: Name of the data array as it is in the vtk file

            :returns: DataArray of the vtk data
        """
        vtkArray = self.output.GetCellData().GetArray(name)
        if vtkArray is None:
            raise AttributeError(f"CellData {name} not found in {self.filename}")
        return DataArray(vtkArray)

    def addPointData(self, array, name):
        """ Add `array` as FIELD of name `name` to the dataset

            :param ndarray array: Numpy array representing the data to add on the mesh vertices
            :param str name: dataName of the added data
        """
        vtkDataArray = numpy_to_vtk(array)
        vtkDataArray.SetName(name)
        self.output.GetPointData().AddArray(vtkDataArray)

    def addCellData(self, array, name):
        """ Add `array` as FIELD of name `name` to the dataset

            :param ndarray array: Numpy array representing the data to add on the cells of the mesh
            :param str name: dataName of the added data
        """
        vtkDataArray = numpy_to_vtk(array)
        vtkDataArray.SetName(name)
        self.output.GetCellData().AddArray(vtkDataArray)

    def removePointData(self, name):
        """ Remove FIELD of name `name` from the dataset

            :param str name: dataName of the data to remove
        """
        vtkArray = self.output.GetPointData().GetArray(name)
        if vtkArray is None: raise AttributeError(f"PointData {name} not found in {self.filename}")
        else: self.output.GetPointData().RemoveArray(name)

    def removeCellData(self, name):
        """ Remove FIELD of name `name` from the dataset

            :param str name: dataName of the data to remove
        """
        vtkArray = self.output.GetCellData().GetArray(name)
        if vtkArray is None: raise AttributeError(f"CellData {name} not found in {self.filename}")
        else: self.output.GetCellData().RemoveArray(name)

    def write(self, fname):
        """ Write vtk unstructured grid to file `fname`

            :param str fname: Output filename
        """
        writer = vtk.vtkUnstructuredGridWriter()
        writer.SetFileName(fname)
        writer.SetInputData(self.output)
        writer.Write()




class Cells:
    """ Class providing a list of all cells in the mesh

        :param vtkOutput data: vtk output of the vtk internal mesh reader
    """
    def __init__(self, data):
        self.data = data
        self.size = data.GetNumberOfCells()

    def __getitem__(self, i):
        """ Accessor to cell *i*

            :param int i: Number of the cell
        """
        if i==self.size:
            raise StopIteration
        return Cell(self.data.GetCell(i), i)

    def __len__(self):
        """ Lenght of the cells list, identical to the number of cells
        """
        return self.size

    def __str__(self):
        """ Provides an easy to read overview of the cells list
        """
        s = "\n   ** Class Cells **\n"
        s += "Size : {} \n".format(self.size)
        return s



class Cell:
    """ Base class to represent a vtk cell.

        :param vtkCell cell: Cell extracted from the mesh by the internal vtk mesh reader
        :param int cellID: Vtk identifier of the cell

        .. csv-table:: Static attributes
            :header: Attribute, Type, Description
            :widths: 20, 10, 70

            `nbP`, int, Number of points corresponding to the cell type
            `ids`, str, Name of cell corresponding to its type

        .. csv-table:: Attributes
            :header: Attribute, Type, Description
            :widths: 20, 10, 70

            `type`, int, Type of the cell
            `id`, int, Id of the cell in the whole mesh
            `points`, list, List of points id

    """
    nbP = {1:1, 3:2, 5:3, 10:4, 12:8}
    ids = {1:"Vertex", 3:"Line", 5:"Triangle", 10:"Tetrahedron", 12:"Hexahedron"}

    def __init__(self, cell, cellID):
        self.type = cell.GetCellType()#cellType
        self.id = cellID
        pts = list()
        for i in range(Cell.nbP[self.type]):
            pts.append(cell.GetPointId(i))
        self.points = pts

    def __getitem__(self, i):
        """ Accessor to the ith point id of the cell

            :param int i: Number of the point in the cell

            :returns: Point id in the whole mesh
        """
        return self.points[i]

    def __len__(self):
        """ Number of points of the cell, corresponding to the cell type
        """
        return Cell.nbP[self.type]

    def __str__(self):
        """ Provides an easy to read overview of the cell
        """
        s = "\n   ** Class Cell **\n"
        s += "Type : {} ({})\n".format(Cell.ids[self.type], self.type)
        s += "Id : {}\n".format(self.id)
        s += "Points : {}\n".format(" ".join(str(p) for p in self.points))
        return s



class DataArray:
    """ Base class to represent a vtk data array

        :param vtkDataArray vtkArray: Vtk data extracted from the mesh by the intenal vtk mesh reader

        .. csv-table:: Attributes
            :header: Attribute, Type, Description
            :widths: 20, 10, 70

            name, str, Name of the array as it is in the vtk mesh
            range, list(2), Range of the data
            size, int, Number of data contained in the array
            componentSize, int, Number of coordinates each data contains
    """
    def __init__(self, vtkArray):
        self.vtkData = vtkArray
        self.name = vtkArray.GetName()
        self.range = vtkArray.GetRange()
        self.size = vtkArray.GetNumberOfTuples()
        self.componentSize = vtkArray.GetNumberOfComponents()

    def __getitem__(self, i):
        """ Accessor to the data in the data array

            :param int i: Number of the data in the array

            :returns: The data contained in the array
        """
        if i==self.size:
            raise StopIteration
        return self.vtkData.GetTuple(i)[0]

    def __str__(self):
        """ Provides an easy to read overview of the data
        """
        s = "\n   ** Class DataArray **\n"
        s += "Name : {} \n".format(self.name)
        s += "Range : {}\n".format(self.range)
        s += "Size : {}\n".format(self.size)
        s += "Component size : {}\n".format(self.componentSize)
        return s

    def __len__(self):
        """ Size of the data array
        """
        return self.size
